from argparse import Namespace
from typing import Optional, TypedDict

from remap_badblocks.src.devices.device_config import DeviceConfig
from remap_badblocks.src.devices.devices_config import DevicesConfig


class SetArgs(TypedDict):
    device_id: Optional[int]
    device_name: Optional[str]
    enable: Optional[bool]
    rename: Optional[str]


def get_device_from_args(args: SetArgs, dc: DevicesConfig) -> DeviceConfig:
    if args["device_id"] is None:
        if args["device_name"] is not None:
            return dc.get_device(name=args["device_name"])
        else:
            raise ValueError(
                "Either --id or --name must be provided to set a device in the database."
            )

    return dc.get_device(id=args["device_id"])


def parse_args(args: Namespace) -> SetArgs:
    return {
        "device_id": args.id,
        "device_name": args.name,
        "enable": args.enable,
        "rename": args.rename,
    }


def set(dc: DevicesConfig, _args: Namespace) -> None:
    """
    Set the configuration of a device from the database.
    """
    args = parse_args(_args)

    device: DeviceConfig = get_device_from_args(args, dc)

    if args["enable"] is not None:
        dc.update_device(
            id=device.id,
            enabled=args["enable"],
        )

    if args["rename"] is not None:
        dc.update_device(
            id=device.id,
            name=args["rename"],
        )

    print("Device config updated.")



import re
from typing import Optional


def parse_string_with_unit_to_bytes(txt: str, unit_multiplier: int) -> int:
    return round(float(txt) * unit_multiplier)


def parse_bytes_to_sectors(_input: int, sector_size: int) -> int:
    if _input % sector_size != 0:
        raise ValueError(
            f"{_input}B is not a multiple of the sector size {sector_size}B."
        )
    return int(_input / sector_size)


def parse_memory_number_to_bytes(txt: str, sector_size: int) -> int:
    """
    Parses a memory value (a sector number or a space size) and outputs a sector number.
    """
    txt = txt.strip()

    try:
        return parse_string_with_unit_to_bytes(txt, sector_size)
    except ValueError:
        pass
    try:
        if txt.endswith("MB"):
            return parse_string_with_unit_to_bytes(txt[:-2], 1024**2)
        elif txt.endswith("GB"):
            return parse_string_with_unit_to_bytes(txt[:-2], 1024**3)
        elif txt.endswith("KB"):
            return parse_string_with_unit_to_bytes(txt[:-2], 1024)
        elif txt.endswith("B"):
            return parse_string_with_unit_to_bytes(txt[:-1], 1)
        else:
            raise ValueError(f"Invalid format: {txt}")
    except ValueError as e:
        raise ValueError(f"Failed to parse memory space from '{txt}': {e}") from e


def parse_memory_range_to_bytes(
    txt: str, sector_size: int
) -> tuple[Optional[int], Optional[int]]:
    """
    Parse a memory space range from a string and check the format is valid. Returns a sector range.
    I.e. checks that the format is 'start-end', where each can be omitted, and start <= end.
    """
    txt = txt.strip()

    m = re.match(
        r"^(?P<start>\d+(\.\d+)?(?:[KMGT]?B)?)?-(?P<end>\d+(\.\d+)?(?:[KMGT]?B)?)?$",
        txt,
    )
    if m is None:
        raise ValueError(
            f"Invalid format: {txt}. Expected format: 'start-end', where each can be omitted."
        )
    groups: dict[str, str | None] = {key: m.group(key) for key in ("start", "end")}
    values: dict[str, int] = {
        key: parse_memory_number_to_bytes(value, sector_size)
        for key, value in groups.items()
        if value is not None
    }

    start, end = values.get("start"), values.get("end")

    if start is not None and end is not None and (end < start):
        raise ValueError(f"End {end} must be greater than or equal to start {start}.")

    return start, end


def parse_memory_range_to_sectors(
    txt: str, sector_size: int
) -> tuple[Optional[int], Optional[int]]:
    start, end = parse_memory_range_to_bytes(txt, sector_size)

    if start is not None:
        start = parse_bytes_to_sectors(start, sector_size)
    if end is not None:
        end = parse_bytes_to_sectors(end, sector_size)

    return start, end



from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Optional, Union

from remap_badblocks.src.badblocks.badblocks import Badblocks
from remap_badblocks.src.mapping import Mapping


@dataclass
class DeviceConfig:
    id: int
    path: Path
    name: str
    sector_size: int
    badblocks: Badblocks
    depends_on: set[int]
    logical_range: tuple[int, int]
    apply_at_startup: bool = False
    spare_sectors: Optional[int] = None
    mapping: Optional[Mapping] = None
    enabled: bool = True

    def __init__(
        self,
        id: int,
        path: Union[Path, str],
        name: str,
        sector_size: int,
        badblocks: Iterable[int],
        logical_range: tuple[int, int],
        apply_at_startup: bool = False,
        depends_on: Iterable[int] = set(),
        spare_sectors: Optional[int] = None,
        mapping: Optional[Mapping] = None,
        enabled: bool = True,
    ):
        if isinstance(path, str):
            path = Path(path)
        self.id = id
        self.path = path
        self.name = name
        self.sector_size = sector_size
        self.badblocks = Badblocks(badblocks)
        self.mapping = mapping if mapping is not None else None
        self.spare_sectors = spare_sectors
        self.depends_on = set(depends_on)
        self.logical_range = logical_range
        self.apply_at_startup = apply_at_startup
        self.enabled = enabled

    def __str__(self) -> str:
        return (
            "DeviceConfig("
            f"   id={self.id}, path={self.path}, name={self.name},"
            f"   sector_size={self.sector_size}, badblocks={len(self.badblocks)}, spare_sectors={self.spare_sectors},"
            f"   mapping={'Undefined' if self.mapping is None else 'Defined'},"
            f"   depends_on={self.depends_on}, logical_range={self.logical_range},"
            f"   {'enabled' if self.enabled else 'disabled'}"
            ")"
        )

    def get_applied_path(self) -> Path:
        return Path("/dev/mapper/", self.name)

    def __hash__(self) -> int:
        return self.id



from argparse import Namespace
from typing import Optional, TypedDict

from remap_badblocks.src.devices.devices_config import DevicesConfig


class RemoveArgs(TypedDict):
    id: Optional[int]
    name: Optional[str]


def parse_args(args: Namespace) -> RemoveArgs:
    return {
        "id": args.id,
        "name": args.name,
    }


def get_device_id_from_args(dc: DevicesConfig, args: RemoveArgs) -> int:
    id: Optional[int] = args["id"]
    name: Optional[str] = args["name"]

    if id is not None and name is not None:
        raise ValueError(
            "Only one of --id or --name can be provided to remove a device from the database."
        )
    elif id is None and name is None:
        raise ValueError(
            "Either --id or --name must be provided to remove a device from the database."
        )
    elif id is not None:
        return id
    else:
        assert name is not None
        return dc.get_device(name=name).id


def remove(dc: DevicesConfig, _args: Namespace) -> None:
    """
    Remove the configuration of a device from the database.
    """
    args = parse_args(_args)
    id = get_device_id_from_args(dc, args)

    dc.remove_device(id)



import contextlib
import io
import tempfile
import threading
from time import sleep, time
from typing import Iterable, Optional, Union

import pytest

from remap_badblocks.src.utils._run_command import (pipe_lines_to_file,
                                                    run_command_realtime)


def test_run_command_realtime_success():
    """Test run_command_realtime with a successful command."""
    cmd = ["echo", "hello world"]
    output = list(run_command_realtime(cmd))
    assert output == ["hello world"]


def test_run_command_realtime_failure():
    """Test run_command_realtime with a failing command."""
    cmd = ["false"]  # `false` is a command that always exits with a non-zero status
    with pytest.raises(RuntimeError, match="Command failed with code 1"):
        list(run_command_realtime(cmd))


def test_run_command_is_captured_asyncronously():
    """Test that the command output is captured asynchronously."""
    # cmd writes hello world to stdout one line per second
    cmd = ["bash", "-c", "for i in {1..5}; do echo 'hello world $i'; sleep 1; done"]
    time_start = time()
    output: Iterable[str] = run_command_realtime(cmd)
    time_launched = time()

    assert time_launched - time_start < 1, "Command should be launched immediately"
    # Check that the output is captured asynchronously
    for i, line in enumerate(output):
        assert line.startswith("hello world")
        assert (
            time() - time_launched < i + 1
        ), "Output should be captured asynchronously"

    assert time() - time_start > 5, "Command should take at least 5 seconds to complete"


@pytest.mark.parametrize(
    "lines, expected_output",
    [
        ([12345, 67890, 54321, "prova"], "12345\n67890\n54321\nprova\n"),  # Normal case
        (
            ["0 6 linear /dev/sda 0", "6 6 linear /dev/sda 10"],
            "0 6 linear /dev/sda 0\n6 6 linear /dev/sda 10\n",
        ),  # DM table case
        ([""], "\n"),  # Empty input
        ([], ""),  # Empty list
    ],
)
def test_pipe_lines_to_file(lines: list[Union[str, int]], expected_output: str):
    # Create a temporary file for testing
    temp_file = tempfile.NamedTemporaryFile(delete=True)
    temp_file_name = temp_file.name

    # Save lines to the temporary file
    pipe_lines_to_file(lines, temp_file_name)

    # Read the contents of the file and check if it matches the expected output
    with open(temp_file_name, "r") as f:
        content = f.read()
        assert content == expected_output

    # Clean up the temporary file
    temp_file.close()


@pytest.mark.parametrize(
    "test_iter",
    [
        range(10),
        ["a", "b", "1234567", "prova"],
        "aeiou",
    ],
)
def test_pipe_lines_to_file_writes_realtime(test_iter: Iterable[Union[str, int]]):
    def generator(path: str):
        for i in test_iter:
            yield i
            sleep(0.5)
            with open(path, "r") as f:
                txt = f.read()
            assert str(i) in txt

    temp_file = tempfile.NamedTemporaryFile(delete=True)
    temp_file_name = temp_file.name

    lines = generator(temp_file_name)

    pipe_lines_to_file(lines, temp_file_name)

    temp_file.close()


def test_run_command_realtime_no_deadlock_with_lots_of_output():
    # This script prints lots of lines to both stdout and stderr very quickly
    # it fails if it times out

    result: dict[str, str] = {}

    def test_logic():
        test_cmd = [
            "python3",
            "-c",
            (
                "import sys\n"
                "for i in range(1000000):\n"
                "    print(f'stdout {i}')\n"
                "    print(f'stderr {i}', file=sys.stderr)"
            ),
        ]

        # redirect stderr to a local variable
        # to avoid printing to the console
        stderr_capture = io.StringIO()

        with contextlib.redirect_stderr(stderr_capture):
            stdout = "\n".join(run_command_realtime(test_cmd))
            stderr = stderr_capture.getvalue()

        result["stdout"] = stdout
        result["stderr"] = stderr

    # Run the test logic in a separate thread
    thread = threading.Thread(target=test_logic)
    thread.start()
    thread.join(timeout=5)  # Wait for the thread to finish

    assert not thread.is_alive(), "Thread should finish within 5 seconds"

    stdout, stderr = result["stdout"], result["stderr"]

    assert len(stdout) > 70000  # 64K is the average OS pipe size
    assert "stdout" in stdout
    assert "stderr" not in stdout
    assert len(stderr) > 70000  # 64K is the average OS pipe size
    assert "stderr" in stderr
    assert "stdout" not in stderr


@pytest.mark.parametrize(
    "stdin,command,expected_out",
    [
        ("Test string\nTest string", ["cat"], "Test string\nTest string"),
        ("Test string", ["cat"], "Test string"),
        ("Test string\n", ["cat"], "Test string"),
        ("a\n1\nb\n2\nc", ["grep", "b"], "b"),
        ("a\n1\nb\n2\nc", ["grep", "\\d"], "1\n2"),
    ],
)
def test_run_command_realtime_stdin(stdin: str, command: list[str], expected_out: str):
    out = "\n".join(run_command_realtime(command, stdin=stdin))
    assert out == expected_out


@pytest.mark.parametrize(
    "stdin,command",
    [
        ("", ["cat"]),
        (None, ["echo"]),
    ],
)
def test_run_command_realtime_deadlock_with_empty_stdout(
    stdin: Optional[str], command: list[str]
):
    o = list(run_command_realtime(command, stdin=stdin))

    assert (not o) or (o[0] == "")



from typing import Iterable

from remap_badblocks.src.devices.device_config import DeviceConfig


def filter_applied_devices(devices: Iterable[DeviceConfig]):
    def check_applied(device: DeviceConfig) -> bool:
        return device.get_applied_path().is_block_device()

    return filter(check_applied, devices)



import sqlite3
from dataclasses import dataclass
from typing import Collection, Iterable, Iterator

from remap_badblocks.src.utils._iterable_bytes_converter import (
    iterable_from_bytes, iterable_to_bytes)

DEFAULT_INT_LENGTH = 8


@dataclass
class MappingElement:
    start_id_virtual: int
    start_id_real: int
    length: int

    INT_LENGTH: int = DEFAULT_INT_LENGTH
    BYTES_LENGTH: int = 3 * INT_LENGTH

    def __bytes__(self) -> bytes:
        """
        Convert the MappingElement to bytes for storage.
        """
        return iterable_to_bytes(
            (
                self.start_id_virtual,
                self.start_id_real,
                self.length,
            ),
            length=self.INT_LENGTH,
        )

    def __index__(self) -> int:
        """
        Convert the MappingElement to an integer for storage.
        """
        return int.from_bytes(self.__bytes__(), "big")

    @classmethod
    def from_bytes(cls, data: bytes) -> "MappingElement":
        """
        Create a MappingElement from bytes.
        """
        if len(data) != cls.BYTES_LENGTH:
            raise ValueError(f"Data must be exactly {cls.BYTES_LENGTH} bytes long.")
        start_id_virtual, start_id_real, length = iterable_from_bytes(
            data, length=cls.INT_LENGTH
        )
        return cls(start_id_virtual, start_id_real, length)

    def __hash__(self) -> int:
        """
        Hash the MappingElement for use in sets or dictionaries.
        """
        return self.__index__()

    @classmethod
    def from_tuple(cls, _tuple: tuple[int, int, int]) -> "MappingElement":
        """
        Create a MappingElement from a tuple.
        """
        if len(_tuple) != 3:
            raise ValueError("Tuple must contain exactly three elements.")
        return cls(
            start_id_virtual=_tuple[0], start_id_real=_tuple[1], length=_tuple[2]
        )

    def to_tuple(self) -> tuple[int, int, int]:
        return self.start_id_virtual, self.start_id_real, self.length

    def __iter__(self) -> Iterator[int]:
        return iter(self.to_tuple())


@dataclass
class Mapping:
    elements: Collection[MappingElement]

    def __bytes__(self) -> bytes:
        """
        Convert the Mapping to bytes for storage.
        """
        return b"".join(bytes(element) for element in self.elements)

    @classmethod
    def from_bytes(cls, data: bytes) -> "Mapping":
        """
        Create a Mapping from bytes.
        """
        if len(data) % MappingElement.BYTES_LENGTH != 0:
            raise ValueError(
                "Data length is not a multiple of {} bytes.",
                MappingElement.BYTES_LENGTH,
            )
        elements: list[MappingElement] = []
        for i in range(0, len(data), MappingElement.BYTES_LENGTH):
            elements.append(
                MappingElement.from_bytes(data[i : i + MappingElement.BYTES_LENGTH])
            )
        return cls(elements)

    def to_sql_binary(self) -> sqlite3.Binary:
        """
        Convert the Mapping to a binary format suitable for SQLite storage.
        """
        return sqlite3.Binary(bytes(self))

    def __iter__(self) -> Iterator[MappingElement]:
        return iter(self.elements)

    @classmethod
    def from_tuples(cls, mapping_tuples: Iterable[tuple[int, int, int]]) -> "Mapping":
        return cls(set(map(MappingElement.from_tuple, mapping_tuples)))



import os
import re
import sqlite3 as sqlite
from collections import OrderedDict
from pathlib import Path
from typing import (Any, Collection, Iterable, Literal, Optional, Union,
                    overload)

from remap_badblocks.src.badblocks.badblocks import Badblocks
from remap_badblocks.src.mapping import Mapping
from remap_badblocks.src.utils._iterable_bytes_converter import (
    iterable_from_bytes, iterable_to_bytes)

from .device_config import DeviceConfig
from .exceptions import DeviceNotFoundError

DEFAULT_INT_LENGTH = 8
DEVICE_INFO_COLUMNS_TYPES = OrderedDict(
    {
        "id": "INTEGER PRIMARY KEY AUTOINCREMENT",
        "path": "TEXT NOT NULL",
        "name": "TEXT NOT NULL",
        "sector_size": "INTEGER NOT NULL",
        "badblocks": "BLOB NOT NULL",
        "spare_sectors": "INTEGER",
        "mapping": "BLOB",
        "depends_on": "BLOB NOT NULL",
        "logical_range_start": "INTEGER NOT NULL",
        "logical_range_end": "INTEGER NOT NULL",
        "apply_at_startup": "BOOLEAN DEFAULT FALSE NOT NULL",
        "enabled": "BOOLEAN DEFAULT TRUE NOT NULL",
    }
)


class DevicesConfig:
    _path: Path

    def __init__(self, path: Union[Path, str]):
        if isinstance(path, str):
            path = Path(path)
        assert not path.is_dir(), "The provided path must not be a directory."
        self._path = path
        self.__init_table()

    def __init_table(self) -> None:
        """
        Initialize the devices table in the database.
        """
        os.makedirs(os.path.abspath(os.path.dirname(self._path)), exist_ok=True)
        with sqlite.connect(self._path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS devices (
                """
                + ",".join(
                    (
                        col_name + " " + col_type
                        for col_name, col_type in DEVICE_INFO_COLUMNS_TYPES.items()
                    )
                )
                + """
                )
                """
            )
            conn.commit()

    def __check_device_in_db(self, conn: sqlite.Connection, id: int) -> bool:
        """
        Check if a device with the given ID exists in the database.
        """
        cursor = conn.cursor()
        cursor.execute("SELECT id FROM devices WHERE id = ? LIMIT 1", (id,))
        val = cursor.fetchone()
        return val is not None

    def _assert_depends_on(
        self, conn: sqlite.Connection, depends_on: Collection[int]
    ) -> None:
        missing: set[int] = set()
        for _id in depends_on:
            if not self.__check_device_in_db(conn, id=_id):
                missing.add(_id)
        if missing:
            raise DeviceNotFoundError(device_id=missing)

    def __parse_device_from_row(
        self,
        id: int,
        path: Union[Path, str],
        name: str,
        sector_size: int,
        badblocks: bytes,
        spare_sectors: Optional[int],
        mapping: Optional[bytes],
        depends_on: bytes,
        logical_range_start: int,
        logical_range_end: int,
        apply_at_startup: bool,
        enabled: bool,
    ) -> DeviceConfig:
        """
        Parse a device configuration from a database row.
        """
        return DeviceConfig(
            id=id,
            path=Path(path),
            name=name,
            sector_size=sector_size,
            badblocks=Badblocks.from_bytes(badblocks),
            spare_sectors=spare_sectors,
            mapping=Mapping.from_bytes(mapping) if mapping is not None else None,
            depends_on=set(iterable_from_bytes(depends_on, length=DEFAULT_INT_LENGTH)),
            logical_range=(logical_range_start, logical_range_end),
            apply_at_startup=apply_at_startup,
            enabled=enabled,
        )

    def get_devices(self) -> Iterable[DeviceConfig]:
        """
        Get the devices configuration as a DataFrame.
        """
        with sqlite.connect(self._path) as conn:
            data = conn.execute(
                """SELECT """
                + ",".join(DEVICE_INFO_COLUMNS_TYPES.keys())
                + """ FROM devices"""
            ).fetchall()

        data = [
            self.__parse_device_from_row(
                **dict(zip(DEVICE_INFO_COLUMNS_TYPES.keys(), row))
            )
            for row in data
        ]
        return data

    @overload
    def get_device(self, *, id: int, name: Literal[None] = None) -> DeviceConfig: ...
    @overload
    def get_device(self, *, id: Literal[None] = None, name: str) -> DeviceConfig: ...

    def get_device(
        self, *, id: Optional[int] = None, name: Optional[str] = None
    ) -> DeviceConfig:
        if id is not None and name is not None:
            raise ValueError("Only one of id or name can be provided to get a device.")

        with sqlite.connect(self._path) as conn:
            cursor = conn.cursor()
            QUERY = (
                """SELECT """
                + ",".join(DEVICE_INFO_COLUMNS_TYPES.keys())
                + """ FROM devices"""
            )
            if id is not None:
                cursor.execute(QUERY + " WHERE id = ?", (id,))
            elif name is not None:
                cursor.execute(QUERY + " WHERE name = ?", (name,))
            else:
                raise ValueError("Either id or name must be provided to get a device.")
            row = cursor.fetchone()
            if row is None:
                raise DeviceNotFoundError(device_id=id, device_name=name)

        return self.__parse_device_from_row(
            **dict(zip(DEVICE_INFO_COLUMNS_TYPES.keys(), row))
        )

    def __insert_into_devices_table(self, cursor: sqlite.Cursor, row: dict[str, Any]):
        cols = list(row.keys())
        cursor.execute(
            """INSERT INTO devices
                ("""
            + ",".join(cols)
            + """)
            VALUES ("""
            + ",".join(("?",) * len(cols))
            + """)""",
            tuple(row[col] for col in cols),
        )

    def add_device(
        self,
        path: Union[Path, str],
        name: str,
        logical_range: tuple[int, int],
        *,
        sector_size: Optional[int] = None,
        badblocks: Union[set[int], Badblocks] = set(),
        depends_on: Collection[int] = set(),
    ) -> None:
        _path: Path = Path(path)
        assert _path.exists(), f"The provided device {_path} does not exist."

        assert self._validate_name_pattern(name), "Name does not match requirements"
        assert not self._is_name_used(name), f"Name {name} is used by other device."

        badblocks = Badblocks(badblocks)

        with sqlite.connect(self._path) as conn:
            self._assert_depends_on(conn, depends_on)

            cursor = conn.cursor()
            self.__insert_into_devices_table(
                cursor,
                {
                    "path": str(_path),
                    "name": name,
                    "sector_size": sector_size,
                    "badblocks": sqlite.Binary(bytes(badblocks)),
                    "spare_sectors": None,
                    "mapping": None,
                    "depends_on": iterable_to_bytes(
                        depends_on, length=DEFAULT_INT_LENGTH
                    ),
                    "logical_range_start": logical_range[0],
                    "logical_range_end": logical_range[1],
                },
            )
            conn.commit()

    def remove_device(self, id: int) -> None:
        with sqlite.connect(self._path) as conn:
            if not self.__check_device_in_db(conn, id=id):
                raise DeviceNotFoundError(device_id=id)

            for dev in self.get_devices():
                if id in dev.depends_on:
                    raise KeyError(
                        f"Device {dev.id} depends on device {id}, can't remove the latter."
                    )

            cursor = conn.cursor()
            cursor.execute("DELETE FROM devices WHERE id = ?", (id,))
            conn.commit()

    def update_device(
        self,
        id: int,
        *,
        path: Optional[Union[Path, str]] = None,
        name: Optional[str] = None,
        badblocks: Optional[Iterable[int]] = None,
        mapping: Optional[Mapping] = None,
        spare_sectors: Optional[int] = None,
        depends_on: Optional[Collection[int]] = None,
        enabled: Optional[bool] = None,
    ) -> None:
        with sqlite.connect(self._path) as conn:
            if not self.__check_device_in_db(conn, id=id):
                raise DeviceNotFoundError(device_id=id)

            cursor = conn.cursor()
            if path is not None:
                cursor.execute(
                    "UPDATE devices SET path = ? WHERE id = ?", (str(path), id)
                )
            if name is not None:
                assert self._validate_name_pattern(
                    name
                ), "Name does not match requirements"
                assert not self._is_name_used(
                    name
                ), f"Name {name} is used by other device."

                cursor.execute("UPDATE devices SET name = ? WHERE id = ?", (name, id))
            if badblocks is not None:
                badblocks = Badblocks(badblocks)
                cursor.execute(
                    "UPDATE devices SET badblocks = ? WHERE id = ?",
                    (sqlite.Binary(bytes(badblocks)), id),
                )
            if mapping is not None:
                cursor.execute(
                    "UPDATE devices SET mapping = ? WHERE id = ?",
                    (sqlite.Binary(bytes(mapping)), id),
                )
            if spare_sectors is not None:
                cursor.execute(
                    "UPDATE devices SET spare_sectors = ? WHERE id = ?",
                    (spare_sectors, id),
                )
            if depends_on is not None:
                self._assert_depends_on(conn, depends_on)
                cursor.execute(
                    "UPDATE devices SET depends_on = ? WHERE id = ?",
                    (
                        iterable_to_bytes(depends_on, length=DEFAULT_INT_LENGTH),
                        id,
                    ),
                )
            if enabled is not None:
                cursor.execute(
                    "UPDATE devices SET enabled = ? WHERE id = ?",
                    (enabled, id),
                )
            conn.commit()

    def __len__(self) -> int:
        """
        Get the number of devices in the database.
        """
        with sqlite.connect(self._path) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT COUNT(*) FROM devices")
            count = cursor.fetchone()[0]
        return count

    def _validate_name_pattern(self, name: str) -> bool:
        return re.fullmatch(r"[0-9A-Za-z#+\-\.:=@_]+", name) is not None

    def _is_name_used(self, name: str) -> bool:
        try:
            self.get_device(name=name)
            return True
        except DeviceNotFoundError:
            return False



from typing import Iterable

import pytest

from remap_badblocks.src.badblocks._remap_badblocks import (
    filter_out_used_and_bad_spare_sectors, is_sector_used_in_mapping,
    remap_badblocks, simplify_mapping)
from remap_badblocks.src.mapping import Mapping


@pytest.mark.parametrize(
    "mapping_in, mapping_out_expected",
    [
        [[(0, 0, 5), (5, 5, 13)], [(0, 0, 18)]],
        [
            [
                (5, 9, 13),
                (0, 4, 5),
            ],
            [
                (0, 4, 18),
            ],
        ],
        [[(0, 0, 500)], [(0, 0, 500)]],
        [
            [
                (15, 3, 1),
                (16, 4, 1),
                (17, 5, 1),
                (18, 6, 2),
                (20, 8, 100),
                (4592, 347, 10),
                (4602, 357, 4),
            ],
            [
                (15, 3, 105),
                (4592, 347, 14),
            ],
        ],
    ],
)
def test_simplify_mapping(
    mapping_in: list[tuple[int, int, int]],
    mapping_out_expected: Iterable[tuple[int, int, int]],
):
    mapping_out = simplify_mapping(mapping_in)

    assert set(mapping_out) == set(mapping_out_expected)


@pytest.mark.parametrize("spare_sectors_n", [10, 20, 100])
@pytest.mark.parametrize(
    "_mapping, badblocks",
    [
        ([(0, 1024, 1024), (1024, 3000, 1000)], [0, 1025, 2050, 2800, 1026]),
        ([(0, 1024, 1024)], [1, 1024, 1025, 1026, 1027, 1028]),
    ],
)
def test_remap_badblocks__no_badblocks_in_output(
    _mapping: Iterable[tuple[int, int, int]],
    badblocks: list[int],
    spare_sectors_n: int,
):
    mapping = Mapping.from_tuples(_mapping)
    spare_sectors = filter_out_used_and_bad_spare_sectors(
        range(spare_sectors_n), mapping, badblocks
    )
    remapped = list(remap_badblocks(mapping, badblocks, spare_sectors))

    for _, start, length in remapped:
        for badblock in badblocks:
            assert not (start <= badblock < start + length)


@pytest.mark.parametrize(
    "mapping, badblocks, spare_sectors_n, is_enough",
    [
        (
            [(0, 1024, 1024)],
            list(range(1024, 1035)),
            10,
            False,
        ),
        (
            [(0, 1024, 1024)],
            list(range(1024, 1034)),
            10,
            True,
        ),
        (
            [(0, 1024, 1024)],
            [],
            10,
            True,
        ),
        [
            [(0, 1024, 1024)],
            [2047],
            0,
            False,
        ],
        [
            [(0, 1024, 1024)],
            list(range(10)) + [2048],
            10,
            True,
        ],
    ],
)
def test_remap_badblocks__not_enough_spares(
    mapping: Iterable[tuple[int, int, int]],
    badblocks: Iterable[int],
    spare_sectors_n: int,
    is_enough: bool,
):
    spare_sectors = iter(range(spare_sectors_n))

    if is_enough:
        list(remap_badblocks(mapping, badblocks, spare_sectors))
    else:
        with pytest.raises(RuntimeError):
            list(remap_badblocks(mapping, badblocks, spare_sectors))


@pytest.mark.parametrize("spare_sectors_n", [10, 20, 100])
@pytest.mark.parametrize(
    "mapping, badblocks",
    [
        ([(0, 1024, 1024), (1024, 3000, 1000)], [0, 1025, 2050, 2800, 1026]),
        ([(0, 1024, 1024)], [1, 1024, 1025, 1026, 1027, 1028]),
        ([(0, 1024, 1024)], [3, 1024, 1025, 1026, 1027, 1028, 2046, 2047, 2048]),
        ([(0, 1024, 1024)], [5, 1024, 1025, 1026, 1027, 1028, 2046, 2047]),
    ],
)
def test_remap_badblocks__is_keeping_good_ranges_constant(
    mapping: Iterable[tuple[int, int, int]],
    badblocks: Iterable[int],
    spare_sectors_n: int,
):
    spare_sectors = iter(range(spare_sectors_n))
    remapped = list(remap_badblocks(mapping, badblocks, spare_sectors))

    old_maps_to_keep: set[tuple[int, int]] = set()

    for start_virt_old, start_real_old, length_old in mapping:
        for virt_old, real_old in zip(
            range(start_virt_old, start_virt_old + length_old),
            range(start_real_old, start_real_old + length_old),
        ):
            if real_old not in badblocks:
                old_maps_to_keep.add((virt_old, real_old))

    new_maps: set[tuple[int, int]] = set()
    for start_virt_new, start_real_new, length_new in remapped:
        for virt_new, real_new in zip(
            range(start_virt_new, start_virt_new + length_new),
            range(start_real_new, start_real_new + length_new),
        ):
            new_maps.add((virt_new, real_new))

    assert old_maps_to_keep.issubset(new_maps)


# @pytest.mark.parametrize(
#     "mapping, n_spare_sectors, answer",
#     [
#         (
#             [(0, 10, 5), (5, 17, 8)],
#             10,
#             0,
#         ),
#         (
#             [(0, 10, 5), (5, 0, 2), (7, 17, 8)],
#             10,
#             2,
#         ),
#         (
#             [(0, 10, 5), (5, 0, 2), (7, 17, 8), (15, 2, 4)],
#             10,
#             6,
#         ),
#     ],
# )
# def test_count_used_spare_sectors(
#     mapping: Iterable[tuple[int, int, int]], n_spare_sectors: int, answer: int
# ):
#     spare_sectors = iter_all_spare_sectors(n_spare_sectors, 0)
#     assert answer == count_used_spare_sectors(mapping, spare_sectors)


@pytest.mark.parametrize("mapping_elem_0", [v * 113 for v in range(100)])
@pytest.mark.parametrize(
    ("sector", "mapping_elem_1_2", "ground_truth"),
    (
        (1, [(0, 4)], True),
        (0, [(0, 4)], True),
        (0, [(1, 4)], False),
        (4, [(0, 4)], False),
        (5, [(0, 4), (5, 1)], True),
        (5, [(0, 4), (6, 5)], False),
        (1000, [(0, 4), (5, 6)], False),
    ),
)
def test_is_sector_used_in_mapping(
    mapping_elem_0: int,
    sector: int,
    mapping_elem_1_2: Iterable[tuple[int, int]],
    ground_truth: bool,
):
    mapping = [
        (mapping_elem_0, *_mapping_elem_1_2) for _mapping_elem_1_2 in mapping_elem_1_2
    ]
    assert is_sector_used_in_mapping(sector, mapping) == ground_truth


@pytest.mark.parametrize(
    ("spare_sectors", "_current_mapping", "badblocks", "ground_truth"),
    (
        ([1, 3, 4, 5], [(0, 5, 2)], {0, 1, 10}, {3, 4}),
        ([], [], [], []),
        ([0, 1, 2], [], [], {0, 1, 2}),
        ([0, 1, 2], [], {0, 1, 2}, []),
        ([0, 1, 2], [(0, 0, 3)], [], []),
        ([], [], {1, 2, 3}, []),
        ([0, 1, 2], [(0, 0, 3)], {0, 1, 2}, []),
    ),
)
def test_filter_out_used_and_bad_spare_sectors(
    spare_sectors: Iterable[int],
    _current_mapping: Iterable[tuple[int, int, int]],
    badblocks: Iterable[int],
    ground_truth: Iterable[int],
):
    current_mapping = Mapping.from_tuples(_current_mapping)
    assert set(
        filter_out_used_and_bad_spare_sectors(
            spare_sectors, current_mapping, set(badblocks)
        )
    ) == set(ground_truth)



from typing import Optional

import pytest

from remap_badblocks.src.utils._parse_inputs import (
    parse_bytes_to_sectors, parse_memory_number_to_bytes,
    parse_memory_range_to_bytes)


@pytest.mark.parametrize(
    "sector_size",
    [512, 1024, 4096],
)
@pytest.mark.parametrize(
    "input_str, expected_output",
    [
        ("0B", 0),
        ("1B", 1),
        ("1024B", 1024),
        ("1KB", 1024),
        ("1MB", 1024 * 1024),
        ("1GB", 1024 * 1024 * 1024),
        ("10MB", 10 * 1024 * 1024),
        ("10GB", 10 * 1024 * 1024 * 1024),
    ],
)
def test_parse_memory_number_to_bytes__no_sectors(
    input_str: str, expected_output: int, sector_size: int
):
    """
    Test the parse_memory_space function with various inputs.
    """
    result = parse_memory_number_to_bytes(input_str, sector_size=sector_size)
    assert result == expected_output, f"Expected {expected_output} but got {result}"


@pytest.mark.parametrize(
    "input_str, sector_size, expected_output",
    [
        ("0", 512, 0),
        ("1", 512, 512),
        ("1", 1024, 1024),
        ("1", 4096, 4096),
        ("1024", 1, 1024),
        ("10", 4096, 40960),
        ("3000000", 1024, 3000000 * 1024),
        ("30000000", 512, 30000000 * 512),
    ],
)
def test_parse_memory_space_sectors(
    input_str: str, sector_size: int, expected_output: int
):
    result = parse_memory_number_to_bytes(input_str, sector_size=sector_size)
    assert result == expected_output, f"Expected {expected_output} but got {result}"


@pytest.mark.parametrize(
    "input_str, sector_size, expected_output",
    [
        ("1-2", 512, (512, 1024)),
        ("1-3", 1024, (1024, 3072)),
        ("1-4", 4096, (4096, 16384)),
        ("1000-2000", 512, (512000, 1024000)),
        ("5MB-6MB", 512, (5 * 1024**2, 6 * 1024**2)),
        ("1000-", 512, (512000, None)),
        ("0-", 512, (0, None)),
        ("-2000", 512, (None, 1024000)),
        ("-2000B", 512, (None, 2000)),
        ("-", 512, (None, None)),
        ("1000-2000KB", 512, (512000, 2048000)),
        ("1000B-2000B", 512, (1000, 2000)),
        ("1.5KB-", 512, (1024 + 512, None)),
    ],
)
def test_parse_memory_range_to_bytes(
    input_str: str, sector_size: int, expected_output: tuple[int, Optional[int]]
):
    """
    Test the parse_memory_range_to_bytes function with various inputs.
    """
    start, end = parse_memory_range_to_bytes(input_str, sector_size=sector_size)
    assert (
        start,
        end,
    ) == expected_output, f"Expected {expected_output} but got {(start, end)}"


@pytest.mark.parametrize(
    "sector_size",
    [512, 1024, 4096],
)
@pytest.mark.parametrize(
    "input_str",
    [
        "",
        "invalid",
        "1-2-3",
        "1-2-3-4",
        "13WB-",
        "-13WB",
        "5-4",
        "5MB-4MB",
        "MB-",
        "MB-MB",
    ],
)
def test_parse_memory_range_to_bytes_breaks(input_str: str, sector_size: int):
    """
    Test the parse_memory_range_to_bytes function breaks with wrong inputs.
    """
    with pytest.raises(ValueError):
        parse_memory_range_to_bytes(input_str, sector_size=sector_size)


@pytest.mark.parametrize(
    "_bytes, sector_size, expected_output",
    [
        (1024, 512, 2),
        (0, 512, 0),
        (512, 512, 1),
    ],
)
def test_parse_bytes_to_sectors(_bytes: int, sector_size: int, expected_output: int):
    assert parse_bytes_to_sectors(_bytes, sector_size) == expected_output


@pytest.mark.parametrize(
    "_bytes, sector_size",
    [
        (1023, 512),
        (1, 512),
        (513, 512),
    ],
)
def test_parse_bytes_to_sectors_errors(_bytes: int, sector_size: int):
    with pytest.raises(ValueError):
        parse_bytes_to_sectors(_bytes, sector_size)



import os
from tomllib import load

from remap_badblocks import __version__


def test_version():
    pyproject = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "..", "pyproject.toml"
    )
    with open(pyproject, "rb") as f:
        assert str(__version__) == load(f)["project"]["version"]



from typing import Iterable


def iterable_to_bytes(iterable: Iterable[int], length: int = 4) -> bytes:
    """
    Convert an iterable of integers to bytes.
    """
    return b"".join(i.to_bytes(length, "big") for i in iterable)


def iterable_from_bytes(data: bytes, length: int = 4) -> Iterable[int]:
    """
    Convert bytes to an iterable of integers.
    """
    if len(data) % length != 0:
        raise ValueError(f"Data length must be a multiple of {length} bytes.")
    return (
        int.from_bytes(data[i : i + length], "big") for i in range(0, len(data), length)
    )



from argparse import Namespace
from typing import Iterable, Optional, TypedDict

from remap_badblocks.src.devices.device_config import DeviceConfig
from remap_badblocks.src.devices.devices_config import DevicesConfig


class GetArgs(TypedDict):
    device_id: Optional[int]
    device_name: Optional[str]
    all: bool


def parse_args(args: Namespace) -> GetArgs:
    return {
        "device_id": args.id,
        "device_name": args.name,
        "all": args.all,
    }


def get_devices_to_print(args: GetArgs, dc: DevicesConfig) -> Iterable[DeviceConfig]:
    if args["device_id"] is not None:
        return [dc.get_device(id=args["device_id"])]
    elif args["device_name"] is not None:
        return [dc.get_device(name=args["device_name"])]
    elif args["all"]:
        return dc.get_devices()
    else:
        return filter(lambda d: d.enabled, dc.get_devices())


def get(dc: DevicesConfig, _args: Namespace) -> None:
    """
    Get the configuration of a device from the database.
    """
    args = parse_args(_args)

    for device in get_devices_to_print(args, dc):
        print(str(device))



import random
from itertools import tee
from typing import Iterable, Iterator

import pytest

from remap_badblocks.src.badblocks._compute_good_ranges import (
    compute_good_ranges, reserve_space_from_good_ranges)
from remap_badblocks.src.remappers._generate_dm_table import generate_dm_table
from remap_badblocks.src.test_utils import (count_sectors_in_ranges,
                                            iter_to_shuffled_generator)


@pytest.mark.parametrize(
    "bad_sectors, available_range, expected_ranges",
    [
        # Basic functionality
        ([2, 3, 5], (0, 9), [(0, 2), (4, 5), (6, 9)]),
        # No bad sectors
        ([], (0, 9), [(0, 9)]),
        ([], (3, 10), [(3, 10)]),
        # All sectors are bad
        (range(10), (0, 10), []),
        # Bad sectors at the beginning and end
        ([0, 1, 8, 9], (0, 10), [(2, 8)]),
        # Ephemeral shuffled iterable
        (iter_to_shuffled_generator(range(10)), (0, 10), []),
        # No broken extremes
        ([2, 3, 5], (0, 10), [(0, 2), (4, 5), (6, 10)]),
        # Uncomplete device
        ([2, 3, 5], (3, 11), [(4, 5), (6, 11)]),
        ([7, 8, 9], (3, 11), [(3, 7), (10, 11)]),
    ],
)
def test_compute_good_ranges(
    bad_sectors: Iterable[int],
    available_range: tuple[int, int],
    expected_ranges: list[tuple[int, int]],
):
    result = list(compute_good_ranges(bad_sectors, available_range=available_range))
    assert result == expected_ranges


@pytest.mark.parametrize(
    "good_ranges, spare_sectors",
    [
        # Test case 1: Basic functionality
        ([(0, 2), (4, 5), (6, 10)], 2),
        # Test case 2: No good ranges
        ([], 2),
        # Test case 3: All good sectors reserved
        ([(0, 10)], 10),
        # Test case 4: Spare sectors exceed available good sectors
        ([(0, 2), (4, 5), (6, 10)], 10),
        # Test case 5: Ephemeral shuffled iterable
        (iter_to_shuffled_generator([(0, 2), (4, 5), (6, 10)]), 10),
        # Test case 6: Ephemeral shuffled iterable
        (iter_to_shuffled_generator([(0, 2), (4, 5), (6, 10)]), 2),
        # Test case 7: Unsorted list
        ([(4, 5), (6, 10), (0, 2)], 2),
        # Test case 8: Real case
        (
            [
                (0, 464992),
                (464994, 825480),
                (825488, 2384152),
                (2384160, 6702752),
                (6702760, 6702768),
                (6702776, 6702816),
                (6702824, 18360728),
                (18360736, 18449304),
                (18449305, 18456312),
                (18456320, 18456328),
                (18456336, 18456344),
                (18456352, 25501304),
                (25501312, 34432448),
                (34432454, 43016120),
                (43016136, 235886992),
                (235886994, 653916408),
                (653916416, 766445288),
                (766445289, 796938504),
                (796938512, 976773168),
            ],
            1048576,
        ),
    ],
)
def test_reserve_space_from_good_ranges(
    good_ranges: Iterator[tuple[int, int]], spare_sectors: int
):
    good_ranges, good_ranges_copy = tee(good_ranges, 2)

    result = list(reserve_space_from_good_ranges(good_ranges, spare_sectors))

    n_original_sectors = count_sectors_in_ranges(good_ranges_copy)
    n_sectors_in_result = count_sectors_in_ranges(result)

    if n_original_sectors >= spare_sectors:
        assert n_original_sectors - n_sectors_in_result == spare_sectors
    else:
        assert n_sectors_in_result == 0


SAMPLE_MAPPING = [
    (1335566, 2384160, 4318592),
    (24452653, 25501312, 8931136),
    (33383789, 34432454, 8583666),
    (795889812, 796938512, 179834656),
    (17407701, 18456352, 7044952),
    (17400678, 18449305, 7007),
    (5654166, 6702776, 40),
    (5654158, 6702760, 8),
    (17407685, 18456320, 8),
    (17407693, 18456336, 8),
    (41967455, 43016136, 192870856),
    (652867725, 653916416, 112528872),
    (765396597, 766445289, 30493215),
    (5654206, 6702824, 11657904),
    (234838311, 235886994, 418029414),
    (17312110, 18360736, 88568),
    (0, 1048586, 1335566),
]


@pytest.mark.parametrize(
    "seed",
    range(10),
)
def test_generate_dm_table_is_sorted(seed: int):
    last_start = -1
    random.seed(seed)
    random.shuffle(SAMPLE_MAPPING)
    for row in generate_dm_table("/dev/test_path", SAMPLE_MAPPING, 512):
        if row:
            row_virtual_start = int(row.split(" ")[0].strip())
            assert row_virtual_start > last_start
            last_start = row_virtual_start



from copy import copy
from typing import Iterable, MutableSequence

from remap_badblocks.src.devices.devices_config import DeviceConfig


def sort_devices_by_dependencies(
    devices: MutableSequence[DeviceConfig],
    already_applied_devices: Iterable[DeviceConfig] = set(),
) -> Iterable[DeviceConfig]:
    sorted_devices: list[DeviceConfig] = []
    already_sorted_ids: set[int] = set(map(lambda x: x.id, already_applied_devices))

    devices = copy(devices)

    while devices:
        for device in devices:
            if all(dep in already_sorted_ids for dep in device.depends_on):
                sorted_devices.append(device)
                already_sorted_ids.add(device.id)
                devices.remove(device)
                break
        else:
            raise ValueError(
                "Circular dependency detected or no device can be applied."
            )

    return sorted_devices



from .add import add
from .apply import apply
from .get import get
from .remove import remove
from .set import set
from .update import update
from .version import version

__all__ = ["add", "apply", "get", "remove", "set", "update", "version"]



from argparse import Namespace
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterable, Literal, Optional, TypedDict

from remap_badblocks.src.badblocks._compute_good_ranges import (
    compute_good_ranges, reserve_space_from_good_ranges)
from remap_badblocks.src.badblocks._find_badblocks import (
    get_all_badblocks, read_known_badblocks)
from remap_badblocks.src.badblocks._mapping_generation import generate_mapping
from remap_badblocks.src.badblocks._remap_badblocks import (
    filter_out_used_and_bad_spare_sectors, iter_all_spare_sectors,
    remap_badblocks, simplify_mapping)
from remap_badblocks.src.devices.device_config import DeviceConfig
from remap_badblocks.src.devices.devices_config import DevicesConfig
from remap_badblocks.src.mapping import Mapping, MappingElement
from remap_badblocks.src.utils._parse_inputs import (
    parse_bytes_to_sectors, parse_memory_number_to_bytes,
    parse_memory_range_to_sectors)
from remap_badblocks.src.utils._run_command import pipe_lines_to_file


class UpdateArgs(TypedDict):
    device_id: Optional[int]
    device_name: Optional[str]
    mode: Literal["read", "write", "skip"]
    output: Optional[Path]
    block_range: str
    spare_space: Optional[str]
    reset_mapping: bool
    external_known_badblocks_file: Optional[Path]


def parse_args(args: Namespace) -> UpdateArgs:
    return {
        "device_id": args.id,
        "device_name": args.name,
        "mode": args.mode,
        "output": args.output,
        "block_range": args.block_range,
        "spare_space": (None if args.spare_space is None else args.spare_space.strip()),
        "reset_mapping": args.reset_mapping,
        "external_known_badblocks_file": args.known_badblocks_file,
    }


def get_device_from_args(args: UpdateArgs, dc: DevicesConfig) -> DeviceConfig:
    if args["device_id"] is None:
        if args["device_name"] is not None:
            return dc.get_device(name=args["device_name"])
        else:
            raise ValueError(
                "Either --id or --name must be provided to update a device in the database."
            )

    return dc.get_device(id=args["device_id"])


def get_update_block_range_from_args(
    args: UpdateArgs, device: DeviceConfig
) -> tuple[int, int]:
    start_block, end_block = parse_memory_range_to_sectors(
        args["block_range"],
        sector_size=device.sector_size,
    )

    if (start_block is not None and start_block < device.logical_range[0]) or (
        end_block is not None and end_block > device.logical_range[1]
    ):
        raise ValueError(
            f"Block range {start_block}-{end_block} must be contained"
            f" within the logical range {device.logical_range[0]}-{device.logical_range[1]}."
        )

    if start_block is None:
        start_block = device.logical_range[0]
    if end_block is None:
        end_block = device.logical_range[1]

    return start_block, end_block


def get_n_spare_sectors_from_args(args: UpdateArgs, device: DeviceConfig) -> int:
    sector_size = device.sector_size

    if args["spare_space"] is None:
        spare_blocks = None
    else:
        spare_blocks = parse_bytes_to_sectors(
            parse_memory_number_to_bytes(args["spare_space"], sector_size=sector_size),
            sector_size,
        )

    if spare_blocks is not None:
        return spare_blocks
    else:
        if device.spare_sectors is None:
            raise ValueError(
                "Should specify spare_sectors the first time you map badblocks"
            )
        return device.spare_sectors


def get_starting_badblocks(args: UpdateArgs, device: DeviceConfig) -> set[int]:
    known_badblocks: set[int] = set(device.badblocks)

    # If known badblocks file is passed, merge it into our known_badblocks
    if args["external_known_badblocks_file"]:
        known_badblocks.update(
            read_known_badblocks(
                known_badblocks_file=args["external_known_badblocks_file"]
            )
        )

    return known_badblocks


def get_current_badblocks(
    args: UpdateArgs,
    known_badblocks: set[int],
    device: DeviceConfig,
    block_range: tuple[int, int],
) -> set[int]:
    if args["mode"] == "skip":
        return known_badblocks

    with NamedTemporaryFile() as known_badblocks_file:
        known_badblocks_file.write("\n".join(map(str, known_badblocks)).encode("utf-8"))
        known_badblocks_file.flush()

        badblocks: set[int] = set(
            get_all_badblocks(
                device.path,
                device.sector_size,
                mode=args["mode"],
                known_badblocks_file=Path(known_badblocks_file.name),
                block_range=block_range,
            )
        )

        if args["output"]:
            pipe_lines_to_file(badblocks, args["output"])

        return badblocks


def update_mapping(
    args: UpdateArgs,
    device: DeviceConfig,
    badblocks: Iterable[int],
    n_spare_sectors: int,
) -> Mapping:
    new_mapping: Iterable[tuple[int, int, int]]

    if (device.mapping is None) or (args["reset_mapping"]):
        good_ranges = compute_good_ranges(
            badblocks, available_range=device.logical_range
        )
        good_ranges = reserve_space_from_good_ranges(good_ranges, n_spare_sectors)
        good_ranges = reserve_space_from_good_ranges(
            good_ranges,
            int(100 * 1024 / device.sector_size),
        )
        new_mapping = generate_mapping(
            good_ranges,
        )
    else:
        badblocks_ = set(badblocks)
        spare_sectors = iter_all_spare_sectors(n_spare_sectors, device.logical_range[0])
        spare_sectors = filter_out_used_and_bad_spare_sectors(
            spare_sectors, device.mapping, badblocks_
        )
        new_mapping = remap_badblocks(
            device.mapping,
            badblocks_,
            spare_sectors,
        )

    new_mapping = list(new_mapping)
    assert new_mapping, "Available blocks are not enough"

    new_mapping = simplify_mapping(new_mapping)

    return Mapping(set(map(MappingElement.from_tuple, new_mapping)))


def update(dc: DevicesConfig, _args: Namespace) -> None:
    """
    Update the mapping of a device in the database.
    """
    args = parse_args(_args)

    device: DeviceConfig = get_device_from_args(args, dc)
    block_range = get_update_block_range_from_args(args, device)
    n_spare_sectors = get_n_spare_sectors_from_args(args, device)

    known_badblocks = get_starting_badblocks(args, device)
    badblocks = get_current_badblocks(args, known_badblocks, device, block_range)

    print(
        f"Badblocks computed. Badblocks went from {len(known_badblocks)} to {len(badblocks)}."
    )

    dc.update_device(
        id=device.id,
        badblocks=badblocks,
    )

    print("Badblocks list updated.")

    new_mapping = update_mapping(args, device, badblocks, n_spare_sectors)
    dc.update_device(id=device.id, mapping=new_mapping, spare_sectors=n_spare_sectors)

    print("Mapping updated.")



import tempfile
from typing import Optional

import pytest

from remap_badblocks.src.badblocks._find_badblocks import (
    build_badblocks_command, parse_badblocks, read_known_badblocks)

test_params: list[tuple[list[str], set[int]]] = [
    (["12345\n", "67890\n", "54321\n"], {12345, 67890, 54321}),  # Normal case
    ([""], set()),  # Empty input
    (["12345\n", "abcde\n", "67890\n"], {12345, 67890}),  # Input with non-numeric lines
    (
        ["   12345   \n", "67890\n   "],
        {12345, 67890},
    ),  # Input with leading/trailing whitespace
]


@pytest.mark.parametrize(
    "input_data, expected_output",
    test_params,
)
def test_parse_badblocks(input_data: list[str], expected_output: set[int]):
    assert set(parse_badblocks(input_data)) == expected_output


@pytest.mark.parametrize(
    "device, mode, known_badblocks_file, sector_size, blocks_range, expected_command",
    [
        # Test case 1: Read mode without known badblocks file
        (
            "/dev/sda",
            "read",
            None,
            512,
            None,
            ["badblocks", "-sv", "-b", "512", "/dev/sda"],
        ),
        # Test case 2: Write mode with known badblocks file
        (
            "/dev/sda",
            "write",
            "/path/to/known_badblocks.txt",
            512,
            None,
            [
                "badblocks",
                "-wsv",
                "-b",
                "512",
                "-i",
                "/path/to/known_badblocks.txt",
                "/dev/sda",
            ],
        ),
        # Test case 3: Different sector size
        (
            "/dev/sda",
            "write",
            "/path/to/known_badblocks.txt",
            1024,
            None,
            [
                "badblocks",
                "-wsv",
                "-b",
                "1024",
                "-i",
                "/path/to/known_badblocks.txt",
                "/dev/sda",
            ],
        ),
        # Test case 4: None path to known badblocks file
        (
            "/dev/sda",
            "write",
            None,
            512,
            None,
            ["badblocks", "-wsv", "-b", "512", "/dev/sda"],
        ),
        # Test case 5: Block range specified
        (
            "/dev/sda",
            "read",
            None,
            512,
            (0, 1000),
            ["badblocks", "-sv", "-b", "512", "/dev/sda", "1000", "0"],
        ),
        # Test case 6: Block range and known badblocks specified
        (
            "/dev/sda",
            "read",
            "/path/to/known_badblocks.txt",
            512,
            (0, 1000),
            [
                "badblocks",
                "-sv",
                "-b",
                "512",
                "-i",
                "/path/to/known_badblocks.txt",
                "/dev/sda",
                "1000",
                "0",
            ],
        ),
    ],
)
def test_build_badblocks_command(
    device: str,
    mode: str,
    known_badblocks_file: Optional[str],
    sector_size: int,
    blocks_range: Optional[tuple[int, int]],
    expected_command: list[str],
):
    assert (
        build_badblocks_command(
            device, sector_size, mode, known_badblocks_file, blocks_range
        )
        == expected_command
    )


def test_build_badblocks_command_error():
    with pytest.raises(ValueError):
        build_badblocks_command("/dev/sda", 512, "invalid_mode", None, None)


def test_read_known_badblocks():
    old_badblocks_file_content = "54321\n67890\n"
    expected_output = {67890, 54321}

    # Create temporary files for testing
    with tempfile.NamedTemporaryFile(delete=True) as temp_known_badblocks_file:
        temp_known_badblocks_file.write(old_badblocks_file_content.encode("utf-8"))
        temp_known_badblocks_file.flush()

        assert (
            set(read_known_badblocks(temp_known_badblocks_file.name)) == expected_output
        )



from typing import Generator, Iterable


def generate_mapping(
    ranges: Iterable[tuple[int, int]],
) -> Generator[tuple[int, int, int], None, None]:
    """Generate a mapping from good ranges."""
    offset = 0
    for start, end in ranges:
        length = end - start
        yield (offset, start, length)  # (start_id_virtual, start_id_real, length)
        offset += length



import subprocess
import sys
import threading
from pathlib import Path
from typing import IO, Iterable, Optional, Protocol, Union


class Stringifyiable(Protocol):
    """Protocol for objects that can be converted to a string."""

    def __str__(self) -> str:
        """Return the string representation of the object."""
        ...


def pipe_stderr_to_stream(pipe: IO[str], stream: IO[str]) -> None:
    for line in pipe:
        print(f"[stderr] {line.strip()}", file=stream)


def run_command_realtime(cmd: list[str], stdin: Optional[str] = None) -> Iterable[str]:
    """Run a command and yield its output line by line in real-time."""
    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdin=subprocess.PIPE,
        text=True,
    ) as process:
        if process.stdout is None:
            raise RuntimeError("Failed to capture stdout from the command.")

        if stdin is not None:
            if process.stdin is None:
                raise RuntimeError("Failed to pass stdin to command.")
            else:
                if not stdin.endswith("\n"):
                    stdin += "\n"
                process.stdin.write(stdin)
                process.stdin.close()

        # Start background stderr reader
        stderr_thread: Optional[threading.Thread] = None
        if process.stderr is not None:
            stderr_thread = threading.Thread(
                target=pipe_stderr_to_stream, args=(process.stderr, sys.stderr)
            )
            stderr_thread.start()

        for line in process.stdout:
            yield line.strip()

        process.wait()
        if stderr_thread is not None:
            stderr_thread.join()

        if process.returncode != 0:
            # Capture stderr if needed
            raise RuntimeError(f"Command failed with code {process.returncode}")


def pipe_lines_to_file(
    lines: Iterable[Stringifyiable], output_file: Union[str, Path, IO[str]]
) -> None:
    """Save the lines to the specified output file in real-time."""
    f: Optional[IO[str]] = None
    try:
        if not isinstance(output_file, IO):
            f = open(output_file, "w")
            output_file = f
        for line in lines:
            output_file.write(str(line) + "\n")
            output_file.flush()  # Ensure the content is written to the file immediately
    finally:
        if f is not None:
            f.close()



import os
from argparse import Namespace
from pathlib import Path
from typing import Optional, TypedDict

from remap_badblocks.src.devices.devices_config import DevicesConfig
from remap_badblocks.src.utils._get_device_info import (
    get_disk_block_size, get_disk_number_of_blocks, resolve_device_name)
from remap_badblocks.src.utils._parse_inputs import \
    parse_memory_range_to_sectors


class AddArgs(TypedDict):
    wwn: Optional[str]
    path: Optional[Path]
    name: str
    depends_on_id: set[int]
    depends_on: set[str]
    logical_sector_range: str


def parse_args(args: Namespace) -> AddArgs:
    return {
        "wwn": args.wwn,
        "path": args.path,
        "name": args.name,
        "depends_on_id": set(args.depends_on_id),
        "depends_on": set(args.depends_on),
        "logical_sector_range": args.logical_sector_range.strip(),
    }


def get_path_from_args(args: AddArgs) -> Path:
    wwn = args["wwn"]
    path = args["path"]

    if wwn is not None and path is not None:
        raise ValueError("Only one of --wwn or --path can be provided, not both.")

    if path is not None:
        if isinstance(path, str):
            path = Path(path)
        return path
    elif wwn is not None:
        if not wwn.startswith("/dev/disk/by-id/"):
            wwn = os.path.join("/dev/disk/by-id/", wwn)
        return Path(wwn)
    else:
        raise ValueError("Either --wwn or --path must be provided.")


def get_device_dependencies_from_args(dc: DevicesConfig, args: AddArgs) -> set[int]:
    return args["depends_on_id"] | {
        dc.get_device(name=dep).id for dep in args["depends_on"] if dep
    }


def get_logical_block_range_from_args(
    args: AddArgs, sector_size: int, total_sectors: int
) -> tuple[int, int]:
    logical_start_block, logical_end_block = parse_memory_range_to_sectors(
        args["logical_sector_range"], sector_size
    )

    if logical_start_block is None:
        logical_start_block = 0
    if logical_end_block is None:
        logical_end_block = total_sectors

    return logical_start_block, logical_end_block


def add(dc: DevicesConfig, _args: Namespace) -> None:
    """
    Add a new device to the database.
    """
    args = parse_args(_args)
    path: Path = get_path_from_args(args)

    sector_size = get_disk_block_size(resolve_device_name(path))
    total_sectors = get_disk_number_of_blocks(resolve_device_name(path))

    dc.add_device(
        path=path,
        name=args["name"],
        sector_size=sector_size,
        depends_on=get_device_dependencies_from_args(dc, args),
        logical_range=get_logical_block_range_from_args(
            args, sector_size, total_sectors
        ),
    )



from argparse import Namespace
from pathlib import Path
from typing import Iterable, Literal, Optional, TypedDict

from remap_badblocks.src.devices.device_config import DeviceConfig
from remap_badblocks.src.devices.devices_config import DevicesConfig
from remap_badblocks.src.mapping import Mapping
from remap_badblocks.src.remappers._check_applied_devices import \
    filter_applied_devices
from remap_badblocks.src.remappers._generate_dm_table import generate_dm_table
from remap_badblocks.src.utils._run_command import run_command_realtime
from remap_badblocks.src.utils._sort_devices import \
    sort_devices_by_dependencies


class ApplyArgs(TypedDict):
    device_id: Optional[int]
    device_name: Optional[str]
    method: Literal["device-mapper"]


def parse_args(args: Namespace) -> ApplyArgs:
    return {
        "device_id": args.id,
        "device_name": args.name,
        "method": args.method,
    }


def get_user_wanted_device_list(
    dc: DevicesConfig, args: ApplyArgs
) -> Iterable[DeviceConfig]:
    if args["device_id"] is not None:
        return [dc.get_device(id=args["device_id"])]
    elif args["device_name"] is not None:
        return [dc.get_device(name=args["device_name"])]
    else:
        return filter(lambda d: d.enabled, dc.get_devices())


def plan_apply_devices(
    dc: DevicesConfig, devices: Iterable[DeviceConfig]
) -> Iterable[DeviceConfig]:
    return sort_devices_by_dependencies(
        list(devices), already_applied_devices=filter_applied_devices(dc.get_devices())
    )


def apply_device_mapper(device: DeviceConfig):
    mapping: Optional[Mapping] = device.mapping
    sector_size: int = device.sector_size
    device_path: Path = device.path

    assert mapping is not None, "Call ‘remap_badblocks update‘ before applying"

    device_dmtable_identifier = device_path

    dmtable = "\n".join(
        generate_dm_table(device_dmtable_identifier, mapping, sector_size)
    )

    for _ in run_command_realtime(["dmsetup", "create", device.name], stdin=dmtable):
        pass


def apply(dc: DevicesConfig, _args: Namespace):
    args = parse_args(_args)

    devices: Iterable[DeviceConfig] = get_user_wanted_device_list(dc, args)
    devices = plan_apply_devices(dc, devices)

    # TODO: remove already applied devices

    method = args["method"]
    assert (
        method == "device-mapper"
    ), f"Only ‘device-mapper‘ method is available, found ‘{method}‘"

    for device in devices:
        apply_device_mapper(device)



from typing import Iterable, Optional, Union


class DeviceNotFoundError(KeyError):
    def __init__(
        self,
        *,
        device_id: Optional[Union[int, Iterable[int]]] = None,
        device_name: Optional[Union[str, Iterable[str]]] = None,
    ) -> None:
        if device_id is not None and device_name is not None:
            raise ValueError("Only one of device_id or device_name can be provided.")
        if device_id is not None:
            message = f"Device(s) with id(s) {device_id} does not exist."
            args = (message,)
        elif device_name is not None:
            message = f"Device(s) with name(s) '{device_name}' does not exist."
            args = (message,)
        else:
            message = "Device(s) not found."
            args = (message,)
        super().__init__(*args)



from random import shuffle
from typing import Generator, Iterable, TypeVar

T = TypeVar("T")


def iter_to_shuffled_generator(_iter: Iterable[T]) -> Generator[T, None, None]:
    _iter_list = list(_iter)
    shuffle(_iter_list)
    for v in _iter_list:
        yield v


def count_sectors_in_ranges(ranges: Iterable[tuple[int, int]]):
    n_sectors = 0
    for start, end in ranges:
        n_sectors += end - start
    return n_sectors



from collections.abc import Container, Generator, Iterable, Iterator
from typing import Optional, Union

from remap_badblocks.src.devices.devices_config import Mapping


def remap_badblocks(
    mapping: Union[Iterable[tuple[int, int, int]], Mapping],
    badblocks: Iterable[int],
    usable_spare_sectors: Iterator[int],
) -> Iterable[tuple[int, int, int]]:
    sorted_badblocks = sorted(badblocks)

    for start_virtual, start_real, length in mapping:
        end_real = start_real + length - 1
        sorted_to_remap: list[int] = []
        sorted_to_remap_virtual: list[int] = []
        for badblock in sorted_badblocks:
            if start_real <= badblock <= end_real:
                sorted_to_remap.append(badblock)
        if not sorted_to_remap:
            # nothing to remap
            yield start_virtual, start_real, length
        else:
            current_start_virtual = start_virtual
            current_start_real = start_real
            for badblock in sorted_to_remap:
                if current_start_real < badblock:
                    # pieces to keep
                    yield current_start_virtual, current_start_real, badblock - current_start_real
                step = badblock - current_start_real + 1
                current_start_virtual += step
                current_start_real += step
                sorted_to_remap_virtual.append(current_start_virtual - 1)
            if current_start_real < end_real:
                # pieces to keep
                yield current_start_virtual, current_start_real, end_real - current_start_real + 1
            for virt_to_remap in sorted_to_remap_virtual:
                try:
                    spare_sector = next(usable_spare_sectors)
                except StopIteration:
                    raise RuntimeError(
                        "Not enough spare sectors to remap all new badblocks"
                    )
                # remap
                yield virt_to_remap, spare_sector, 1


def identify_simplifiable_couple_in_mapping(
    mapping: list[tuple[int, int, int]],
) -> Optional[tuple[int, int]]:
    for i, (start_virtual_0, start_real_0, length_0) in enumerate(mapping[:-1]):
        end_virtual_0 = start_virtual_0 + length_0
        end_real_0 = start_real_0 + length_0
        for _j, (start_virtual_1, start_real_1, length_1) in enumerate(
            mapping[i + 1 :]
        ):
            j = _j + i + 1
            end_virtual_1 = start_virtual_1 + length_1
            end_real_1 = start_real_1 + length_1
            if (start_virtual_0 == end_virtual_1) and (start_real_0 == end_real_1):
                return j, i
            if (start_virtual_1 == end_virtual_0) and (start_real_1 == end_real_0):
                return i, j
    return None


def simplify_mapping(
    mapping: list[tuple[int, int, int]],
) -> Iterable[tuple[int, int, int]]:
    while (to_simplify := identify_simplifiable_couple_in_mapping(mapping)) is not None:
        i, j = to_simplify
        element_i = mapping[i]
        element_j = mapping[j]

        simplified_element = element_i[0], element_i[1], element_i[2] + element_j[2]

        _i = min(i, j)
        _j = max(i, j)

        mapping = (
            mapping[:_i]
            + mapping[_i + 1 : _j]
            + mapping[_j + 1 :]
            + [simplified_element]
        )

    return mapping


def iter_all_spare_sectors(
    n_spare_sectors: int, first_spare_sector: int
) -> Iterator[int]:
    return iter(range(first_spare_sector, first_spare_sector + n_spare_sectors))


def is_sector_used_in_mapping(
    s: int, mapping: Union[Iterable[tuple[int, int, int]], Mapping]
):
    for _, start, length in mapping:
        end = start + length
        if start <= s < end:
            return True
    return False


def filter_out_used_and_bad_spare_sectors(
    spare_sectors: Iterable[int],
    current_mapping: Mapping,
    badblocks: Container[int],
) -> Generator[int]:
    for s in spare_sectors:
        if s not in badblocks and not is_sector_used_in_mapping(s, current_mapping):
            yield s



from itertools import chain
from pathlib import Path
from typing import Iterable, Optional, Union

from remap_badblocks.src.utils._run_command import run_command_realtime


def parse_badblocks(lines: Iterable[str]) -> Iterable[int]:
    """Parse badblocks output and return a set of bad sector numbers."""
    lines = map(lambda line: line.strip(), lines)
    lines = filter(lambda line: line.isdigit(), lines)
    return map(int, lines)


def read_known_badblocks(known_badblocks_file: Union[Path, str]) -> Iterable[int]:
    """Merge badblocks with known badblocks from a file."""
    with open(known_badblocks_file, "r") as f:
        known_badblocks = parse_badblocks(f.readlines())
    return known_badblocks


def build_badblocks_command(
    device: Union[Path, str],
    sector_size: int,
    mode: str,
    known_badblocks_file: Optional[Union[Path, str]],
    blocks_range: Optional[tuple[int, int]],
) -> list[str]:
    """
    Build the badblocks command based on the mode and optional known badblocks file.
    blocks_range is a right-open interval
    """
    if mode == "read":
        cmd = ["badblocks", "-sv"]
    elif mode == "write":
        cmd = ["badblocks", "-wsv"]
    else:
        raise ValueError("Invalid mode. Use 'read' or 'write'.")

    cmd += ["-b", str(sector_size)]

    if known_badblocks_file:
        cmd += ["-i", str(known_badblocks_file)]

    cmd += [str(device)]

    if blocks_range:
        start_block, end_block = blocks_range
        cmd += [str(end_block), str(start_block)]

    return cmd


def get_all_badblocks(
    device: Path,
    sector_size: int,
    mode: str = "read",
    known_badblocks_file: Optional[Path] = None,
    block_range: Optional[tuple[int, int]] = None,
) -> Iterable[int]:
    """
    Run badblocks on a device and return the results as a set of bad sectors.
    block_range is a right-open interval
    """
    badblocks: Iterable[int] = set()

    if known_badblocks_file:
        badblocks = read_known_badblocks(known_badblocks_file)

    cmd = build_badblocks_command(
        device, sector_size, mode, known_badblocks_file, block_range
    )
    badblocks_lines = run_command_realtime(cmd)
    badblocks = chain(badblocks, parse_badblocks(badblocks_lines))

    return badblocks



from typing import Generator, Iterable


def compute_good_ranges(
    bad_sectors: Iterable[int], available_range: tuple[int, int]
) -> Generator[tuple[int, int], None, None]:
    """
    Compute good sector ranges, excluding bad sectors.
    available_range is expected to be a range that is right-open
    result is a range that is right-open
    """
    bad_sectors = sorted(set(bad_sectors))  # Ensure bad_sectors is sorted and unique
    current_start = available_range[0]
    last_sector = available_range[1] - 1

    for bad_sector in bad_sectors:
        if bad_sector > last_sector:
            break
        if current_start < bad_sector:
            yield (current_start, bad_sector)
        current_start = bad_sector + 1

    if current_start <= last_sector:
        yield (current_start, last_sector + 1)


def reserve_space_from_good_ranges(
    good_ranges: Iterable[tuple[int, int]], spare_sectors: int
) -> Generator[tuple[int, int], None, None]:
    """
    Return good sector ranges, skipping the first `spare_sectors` good sectors as reserved space.
    Both input and output ranges are right-open.
    """
    to_reserve = spare_sectors
    for start, end in good_ranges:
        length = end - start
        if to_reserve >= length:
            to_reserve -= length
            continue
        elif to_reserve > 0:
            start += to_reserve
            to_reserve = 0
        yield (start, end)



from typing import Iterable

import pytest

from remap_badblocks.src.badblocks._mapping_generation import generate_mapping
from remap_badblocks.src.remappers._generate_dm_table import generate_dm_table


@pytest.mark.parametrize(
    "ranges, block_size, expected_output",
    [
        ([(0, 6), (10, 16)], 512, ["0 6 linear /dev/sda 0", "6 6 linear /dev/sda 10"]),
        (
            iter([(0, 6), (10, 16)]),
            512,
            ["0 6 linear /dev/sda 0", "6 6 linear /dev/sda 10"],
        ),
        (
            iter([(10, 16), (0, 6)]),
            512,
            ["0 6 linear /dev/sda 10", "6 6 linear /dev/sda 0"],
        ),
        (
            [(0, 6), (10, 16)],
            1024,
            ["0 12 linear /dev/sda 0", "12 12 linear /dev/sda 20"],
        ),
    ],
)
def test_generate_dm_table(
    ranges: Iterable[tuple[int, int]], block_size: int, expected_output: list[str]
):
    device = "/dev/sda"

    mapping = generate_mapping(ranges)
    result = list(generate_dm_table(device, mapping, block_size))

    assert len(result) == len(expected_output)

    assert set(result) == set(
        expected_output
    ), f"Expected {expected_output}, but got {result}"



from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Iterable, Optional

import pytest

from remap_badblocks.src.devices.devices_config import DevicesConfig
from remap_badblocks.src.mapping import Mapping, MappingElement
from remap_badblocks.src.utils._iterable_bytes_converter import (
    iterable_from_bytes, iterable_to_bytes)

BADBLOCKS_SAMPLES: list[Optional[set[int]]] = [
    None,
    set(),
    {1, 2, 3},
    {100, 200, 300},
    {
        0,
        10 * (1024**3) * 2,
    },  # Edge cases for memory limits (test 10TB badblock for 512-bytes block size)
]


@pytest.mark.parametrize(
    "start_id_virtual, start_id_real, length",
    [
        (0, 0, 0),
        (1, 2, 3),
        (12345678, 87654321, 1000),
        (4294967295, 4294967295, 4294967295),
    ],
)
def test_mapping_element_bytes(start_id_virtual: int, start_id_real: int, length: int):
    """
    Test the conversion of MappingElement to and from bytes.
    """
    element = MappingElement(start_id_virtual, start_id_real, length)
    data = bytes(element)
    restored_element = MappingElement.from_bytes(data)

    assert restored_element.start_id_virtual == start_id_virtual
    assert restored_element.start_id_real == start_id_real
    assert restored_element.length == length
    assert isinstance(data, bytes)
    assert len(data) == MappingElement.BYTES_LENGTH


@pytest.mark.parametrize(
    "elements",
    [
        [MappingElement(0, 0, 0)],
        [MappingElement(1, 2, 3), MappingElement(4, 5, 6)],
        [MappingElement(12345678, 87654321, 1000)],
        [MappingElement(4294967295, 4294967295, 4294967295)],
    ],
)
def test_mapping_bytes(elements: list[MappingElement]):
    """
    Test the conversion of Mapping to and from bytes.
    """
    mapping = Mapping(elements)
    data = bytes(mapping)
    restored_mapping = Mapping.from_bytes(data)

    original_elements = set(mapping.elements)
    restored_elements = set(restored_mapping.elements)

    assert original_elements == restored_elements
    assert isinstance(data, bytes)
    assert isinstance(restored_mapping, Mapping)
    assert len(data) % MappingElement.BYTES_LENGTH == 0


def test_devices_config_init():
    """
    Test the initialization of DevicesConfig with an empty database.
    """
    with NamedTemporaryFile() as temp_db:
        devices_config = DevicesConfig(path=temp_db.name)
        assert devices_config.get_devices() == []

        with pytest.raises(KeyError):
            devices_config.get_device(id=1)

        assert len(devices_config) == 0


@pytest.mark.parametrize(
    "length",
    [4, 8, 16],
)
@pytest.mark.parametrize(
    "iterable",
    [
        [],
        [1, 2, 3],
        {100, 400, 600},
    ],
)
def test_iterable_to_bytes(length: int, iterable: Iterable[int]):
    assert set(iterable) == set(
        iterable_from_bytes(iterable_to_bytes(iterable, length=length), length=length)
    )


@pytest.mark.parametrize(
    "sector_size",
    [512, 1024, 2048],
)
@pytest.mark.parametrize(
    "badblocks",
    BADBLOCKS_SAMPLES,
)
def test_devices_config_add_device(badblocks: Optional[set[int]], sector_size: int):
    """
    Test adding a device to DevicesConfig.
    """
    with NamedTemporaryFile() as temp_block_file:
        with NamedTemporaryFile() as temp_db:
            devices_config = DevicesConfig(path=temp_db.name)

            if badblocks is None:
                devices_config.add_device(
                    name="name1",
                    path=temp_block_file.name,
                    sector_size=sector_size,
                    logical_range=(0, int(1e10)),
                )
            else:
                devices_config.add_device(
                    name="name1",
                    path=temp_block_file.name,
                    badblocks=badblocks,
                    sector_size=sector_size,
                    logical_range=(0, int(1e10)),
                )

            assert len(devices_config) == 1

            devices = list(devices_config.get_devices())

            device = devices[0]
            assert device.path == Path(temp_block_file.name)
            if badblocks is not None:
                assert len(device.badblocks) == len(badblocks)
                assert set(device.badblocks) == set(badblocks)
            assert device.mapping is None

            devices_config.add_device(
                name="name2",
                path=temp_block_file.name,
                sector_size=sector_size,
                logical_range=(0, int(1e10)),
            )

            assert len(devices_config) == 2

            with pytest.raises(Exception):
                devices_config.add_device(
                    name="name2",
                    path=temp_block_file.name,
                    sector_size=sector_size,
                    logical_range=(0, int(1e10)),
                )


@pytest.mark.parametrize(
    "sector_size",
    [512, 1024, 2048],
)
def test_devices_config_remove_device(sector_size: int):
    """
    Test removing a device from DevicesConfig.
    """
    with NamedTemporaryFile() as temp_block_file:
        with NamedTemporaryFile() as temp_db:
            devices_config = DevicesConfig(path=temp_db.name)

            devices_config.add_device(
                name="name1",
                path=temp_block_file.name,
                sector_size=sector_size,
                logical_range=(0, int(1e10)),
            )
            devices_config.add_device(
                name="name2",
                path=temp_block_file.name,
                sector_size=sector_size,
                logical_range=(0, int(1e10)),
            )
            assert len(devices_config) == 2

            devices = list(devices_config.get_devices())
            device_to_remove_id = devices[0].id
            device_to_keep_id = devices[1].id

            assert device_to_remove_id != device_to_keep_id

            devices_config.remove_device(id=device_to_remove_id)
            assert len(devices_config) == 1

            with pytest.raises(KeyError):
                devices_config.get_device(id=device_to_remove_id)

            devices_config.get_device(id=device_to_keep_id)


def test_devices_config_update_device():
    """
    Test updating a device in DevicesConfig.
    """
    with (
        NamedTemporaryFile() as temp_block_file1,
        NamedTemporaryFile() as temp_block_file2,
    ):
        with NamedTemporaryFile() as temp_db:
            devices_config = DevicesConfig(path=temp_db.name)

            devices_config.add_device(
                name="name",
                path=temp_block_file1.name,
                sector_size=512,
                logical_range=(0, int(1e10)),
            )
            assert len(devices_config) == 1

            devices = list(devices_config.get_devices())
            device_id = devices[0].id

            new_path = Path(temp_block_file2.name)
            devices_config.update_device(id=device_id, path=new_path)

            updated_device = devices_config.get_device(id=device_id)
            assert updated_device.path == new_path

            new_badblocks = {1, 2, 3}
            devices_config.update_device(id=device_id, badblocks=new_badblocks)

            updated_device = devices_config.get_device(id=device_id)
            assert set(updated_device.badblocks) == new_badblocks



import os
import re
from pathlib import Path


def resolve_device_name(device: Path) -> str:
    resolved = device.resolve()
    if not resolved.is_block_device():
        raise ValueError(f"{device} is not a valid block device.")
    _resolved = str(resolved)
    m = re.match(r"^/dev/([a-zA-Z0-9\-]+)$", _resolved)
    if not m:
        raise RuntimeError(f"Could not parse '{_resolved}'.")
    return m.group(1)


def get_disk_block_size(device_name: str) -> int:
    """Get the block size of the disk."""
    path = os.path.join("/sys/block/", device_name, "queue/physical_block_size")
    try:
        with open(path, "r") as f:
            block_size = f.read()
            return int(block_size.strip())
    except Exception as e:
        raise RuntimeError(f"Could not read block size for {device_name}: {e}") from e


def get_disk_number_of_blocks(device_name: str) -> int:
    """Get the number of blocks on the disk."""
    block_size = get_disk_block_size(device_name)
    path = os.path.join("/sys/block/", device_name, "size")
    try:
        with open(path, "r") as f:
            txt = f.read()
        size_in_blocks = int(txt.strip()) * 512 / block_size
        assert (
            size_in_blocks.is_integer()
        ), f"Size in blocks is not an integer: {size_in_blocks}"
        return int(size_in_blocks)
    except Exception as e:
        raise RuntimeError(
            f"Could not read size in physical blocks for {device_name}: {e}"
        ) from e



from dataclasses import dataclass
from typing import Iterable, Iterator

from remap_badblocks.src.utils._iterable_bytes_converter import (
    iterable_from_bytes, iterable_to_bytes)


@dataclass
class Badblocks(Iterable[int]):
    badblocks: set[int]

    INT_LENGTH: int = 8

    def __init__(self, badblocks: Iterable[int]):
        self.badblocks = set(badblocks)

    def __bytes__(self) -> bytes:
        """
        Convert the Badblocks to bytes for storage.
        """
        return iterable_to_bytes(self.badblocks, length=self.INT_LENGTH)

    @classmethod
    def from_bytes(cls, data: bytes) -> "Badblocks":
        """
        Create a Badblocks instance from bytes.
        """
        return cls(badblocks=set(iterable_from_bytes(data, length=cls.INT_LENGTH)))

    def __len__(self) -> int:
        """
        Get the number of bad blocks.
        """
        return len(self.badblocks)

    def __iter__(self) -> Iterator[int]:
        """
        Iterate over the bad blocks.
        """
        return iter(self.badblocks)



from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Callable

from remap_badblocks.cli.commands import (add, apply, get, remove, set, update,
                                          version)
from remap_badblocks.src.devices.devices_config import DevicesConfig
from remap_badblocks.src.devices_config_constants import \
    DEFAULT_DEVICES_CONFIG_PATH

argparse = ArgumentParser(
    description="Badblocks remapper for block devices.",
)

argparse.add_argument(
    "-P",
    "--db-path",
    type=Path,
    default=DEFAULT_DEVICES_CONFIG_PATH,
    help="Path to the devices configuration database.",
)

subparser = argparse.add_subparsers(
    title="Actions",
    description="Available actions for the badblocks remapper.",
    dest="action",
    help="Action to perform",
    required=True,
)

add_new_device_parser = subparser.add_parser(
    "add",
    help="Add a new device to the database.",
)

add_new_device_parser.add_argument(
    "name", type=str, help="The name of the output device."
)
add_new_device_parser.add_argument(
    "--wwn",
    type=str,
    help="The WWN of the device to add.",
)
add_new_device_parser.add_argument(
    "--path",
    type=Path,
    help="The path to the device to add.",
)
add_new_device_parser.add_argument(
    "--depends-on",
    type=str,
    default=[],
    action="append",
    help=(
        "The name of a device that this device depends on. Can be specified multiple times to add multiple"
        " dependencies. This is used to ensure that devices are applied in the correct order."
    ),
)
add_new_device_parser.add_argument(
    "--depends-on-id",
    type=int,
    default=[],
    action="append",
    help=(
        "The ID of a device that this device depends on. Can be specified multiple times to add multiple"
        " dependencies. This is used to ensure that devices are applied in the correct order."
    ),
)
add_new_device_parser.add_argument(
    "--logical-sector-range",
    type=str,
    default="-",
    help=(
        "Logical sector range to assign to the device. Format: 'start-end', where 'end' can be omitted to mean"
        " 'till the end of the device'. Both start and end can be specified as sector numbers or as byte offsets."
        " In case of byte offsets, they must be multiples of the sector size. If not specified, the whole device"
        " will be used. Note that end is not included in the range."
    ),
)


get_devices_parser = subparser.add_parser(
    "get",
    help="Get the existing device(s) in the database.",
)
get_devices_parser.add_argument(
    "--id",
    type=int,
    default=None,
    help="The ID of the device to get. If neither --id or --name are provided, all devices will be returned.",
)
get_devices_parser.add_argument(
    "--name",
    type=str,
    default=None,
    help="The name of the device to get. If neither --id or --name are provided, all devices will be returned.",
)
get_devices_parser.add_argument(
    "-a",
    "--all",
    action="store_true",
    help="Also get disabled devices.",
)

remove_devices_parser = subparser.add_parser(
    "remove",
    help="Remove an existing device in the database.",
)
remove_devices_parser.add_argument(
    "--id",
    type=int,
    default=None,
    help="The ID of the device to remove. If neither --id or --name are provided, an error will be raised.",
)
remove_devices_parser.add_argument(
    "--name",
    type=str,
    default=None,
    help="The name of the device to remove. If neither --id or --name are provided, an error will be raised.",
)

update_mapping_parser = subparser.add_parser(
    "update",
    help=(
        "Update the mapping of a device in the database. More specifically: compute/update the device's badblocks and"
        " store them in the database, compute the new mapping of a device; if the mapping has changed, the user will"
        " be prompted to confirm the changes."
    ),
)
update_mapping_parser.add_argument(
    "--id",
    type=int,
    default=None,
    help="The ID of the device to update.",
)
update_mapping_parser.add_argument(
    "--name",
    type=str,
    default=None,
    help="The name of the device to update.",
)
update_mapping_parser.add_argument(
    "--mode",
    type=str,
    choices=["read", "write", "skip"],
    default="read",
    help="Mode for computing badblocks",
)
update_mapping_parser.add_argument(
    "--known-badblocks-file",
    type=Path,
    default=None,
    help=(
        "Path to file with known badblocks (a sector number for each line) that will be merged to those found so far."
        " Take care the sector size is the same as configured into remap_badblocks"
    ),
)
update_mapping_parser.add_argument(
    "--block-range",
    type=str,
    help=(
        "Block range to check (e.g., 0-1000, or 1573-), omitted start and end means the whole logical range. Note"
        " that the end is not included in the range"
    ),
    default="-",
)
update_mapping_parser.add_argument(
    "--output",
    type=Path,
    default=None,
    help="Badblocks are stored internally. If provided, they will also be copied to this file.",
)
update_mapping_parser.add_argument(
    "--spare-space",
    type=str,
    default=None,
    help="Number of spare sectors to reserve from the good ranges. Cannot be changed after first time",
)
update_mapping_parser.add_argument(
    "--reset-mapping",
    action="store_true",
    help=(
        "If specified, will ignore existing mapping and build a new mapping from scratch."
        " DANGER: this might lead to data loss, handle with care"
    ),
)

apply_devices_parser = subparser.add_parser(
    "apply",
    help=(
        "Apply a remapping that's already existing in the database. All provided devices will be applied,"
        " regardless if it's already applied. This might cause errors."
    ),
)
apply_devices_parser.add_argument(
    "--id",
    type=int,
    default=None,
    help="The ID of the device to apply. If neither --id or --name are provided, all devices will be applied.",
)
apply_devices_parser.add_argument(
    "--name",
    type=str,
    default=None,
    help="The name of the device to apply. If neither --id or --name are provided, all devices will be applied.",
)
apply_devices_parser.add_argument(
    "--method",
    type=str,
    default="device-mapper",
    choices=["device-mapper"],
    help="Method to apply the mapping. Only device-mapper is available",
)

set_device_parser = subparser.add_parser(
    "set",
    help="Set config of an already existing device in the database.",
)
set_device_parser.add_argument(
    "--id",
    type=int,
    default=None,
    help="The ID of the device to set. Either --id or --name must be provided.",
)
set_device_parser.add_argument(
    "--name",
    type=str,
    default=None,
    help="The name of the device to set. Either --id or --name must be provided.",
)
set_device_parser.add_argument(
    "--enable",
    action="store_true",
    help="Enable the device.",
)
set_device_parser.add_argument(
    "--disable",
    dest="enable",
    action="store_false",
    help="Disable the device.",
)
set_device_parser.set_defaults(
    enable=None,
)
set_device_parser.add_argument(
    "--rename",
    type=str,
    help="Rename the device.",
)

version_parser = subparser.add_parser("version")


def main() -> None:
    args = argparse.parse_args()
    db_path: Path = args.db_path

    actions: dict[str, Callable[[DevicesConfig, Namespace], None]] = {
        "add": add,
        "apply": apply,
        "get": get,
        "set": set,
        "remove": remove,
        "update": update,
    }

    static_actions: dict[str, Callable[[], None]] = {
        "version": version,
    }

    if args.action in actions:
        if db_path.is_dir():
            raise ValueError("The provided path must not be a directory.")
        devices_config = DevicesConfig(db_path)
        actions[args.action](devices_config, args)
    elif args.action in static_actions:
        static_actions[args.action]()
    else:
        raise ValueError(
            f"Unknown action: {args.action}. Please use one of {tuple(actions.keys()) + tuple(static_actions.keys())}."
        )


if __name__ == "__main__":
    main()



import random

import pytest

from remap_badblocks.src.devices.devices_config import DeviceConfig
from remap_badblocks.src.utils._sort_devices import \
    sort_devices_by_dependencies

DEVICES_SAMPLE = [
    DeviceConfig(
        id=1,
        depends_on=[],
        name="dev1",
        path="/dev/sda",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=2,
        depends_on=[3],
        name="dev2",
        path="/dev/sdb",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=3,
        depends_on=[1],
        name="dev3",
        path="/dev/sdc",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=3,
        depends_on=[1, 2, 3],
        name="dev4",
        path="/dev/sdd",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
]

DEVICES_SAMPLE_NO_DEPS = [
    DeviceConfig(
        id=1,
        depends_on=[],
        name="dev1",
        path="/dev/sda",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=2,
        depends_on=[],
        name="dev2",
        path="/dev/sdb",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=3,
        depends_on=[],
        name="dev3",
        path="/dev/sdc",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=3,
        depends_on=[],
        name="dev4",
        path="/dev/sdd",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
]

DEVICES_SAMPLE_CIRCULAR_DEP = [
    DeviceConfig(
        id=1,
        depends_on=[],
        name="dev1",
        path="/dev/sda",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=2,
        depends_on=[3],
        name="dev2",
        path="/dev/sdb",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=3,
        depends_on=[2],
        name="dev3",
        path="/dev/sdc",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
    DeviceConfig(
        id=4,
        depends_on=[],
        name="dev4",
        path="/dev/sdd",
        sector_size=512,
        badblocks=[],
        logical_range=(0, 1000),
    ),
]


def test_sort_devices__empty_input():
    devices: list[DeviceConfig] = []
    sorted_devices = list(sort_devices_by_dependencies(devices))
    assert sorted_devices == []


def test_sort_devices__single_device():
    devices = DEVICES_SAMPLE[:1]
    sorted_devices = list(sort_devices_by_dependencies(devices))
    assert sorted_devices == devices


def test_sort_devices__no_dependencies():
    sorted_devices = list(sort_devices_by_dependencies(DEVICES_SAMPLE_NO_DEPS))
    for dev in DEVICES_SAMPLE_NO_DEPS:
        assert dev in sorted_devices
    assert len(sorted_devices) == len(DEVICES_SAMPLE_NO_DEPS)


def test_sort_devices__circular_dependency():
    with pytest.raises(ValueError, match=r"[Cc]ircular [Dd]ependency"):
        list(sort_devices_by_dependencies(DEVICES_SAMPLE_CIRCULAR_DEP))


@pytest.mark.parametrize(
    "seed",
    range(10),
)
def test_sort_devices(seed: int):
    random.seed(seed)
    CURRENT_SAMPLE = DEVICES_SAMPLE.copy()
    random.shuffle(CURRENT_SAMPLE)

    sorted_devices = list(sort_devices_by_dependencies(CURRENT_SAMPLE))

    # Check that all devices are present
    for dev in CURRENT_SAMPLE:
        assert dev in sorted_devices

    # Check that dependencies are respected
    already_there_ids: set[int] = set()

    for dev in sorted_devices:
        assert set(dev.depends_on).issubset(already_there_ids)
        already_there_ids.add(dev.id)



from pathlib import Path
from typing import Generator, Iterable, TypeVar, Union

from remap_badblocks.src.devices.devices_config import Mapping


def generate_dm_table(
    device: Union[Path, str],
    mapping: Union[Iterable[tuple[int, int, int]], Mapping],
    block_size: int,
) -> Generator[str, None, None]:
    """Generate a device-mapper linear mapping table from good ranges."""
    block_size_multiplier = block_size / 512
    assert block_size_multiplier.is_integer(), "Block size must be a multiple of 512"
    block_size_multiplier = int(block_size_multiplier)

    T = TypeVar("T")

    def get_first(iterable: Iterable[T]) -> T:
        return next(iter(iterable))

    for start_virtual, start_real, length in sorted(mapping, key=get_first):
        start_virtual *= block_size_multiplier
        start_real *= block_size_multiplier
        length *= block_size_multiplier

        yield f"{start_virtual} {length} linear {device} {start_real}"
